#ifndef __HIERARCHICAL_CLUSTERING__
#define __HIERARCHICAL_CLUSTERING__

/*! \file HierarchicalClustering.h
 *  \brief Header File for Hierarchical Clustering Algorithms
 *
 * The hierarchical clustering algorithms are used to initialize the EM algorithms
 * to provide good starting points.
 * There are totally 7 hierachical clustering algorithms available in this file.
 */

#include <iostream>
#include <cstdlib>
#include <cmath>

#include "VectorUtilities.h"

using namespace std;

//! Enumeration of Hierarchical Clustering Algorithms
/*! AVE: Average
 *
 *  CEN: Centroid
 *
 *  COM: Complete Linkage
 *
 *  MCQ: McQuitty
 *
 *  MED: Median
 *
 *  SIN: Single Linkage
 *
 *  WAR: Ward's Minimum Variance Method (Recommended for Genetic Data)
 */
enum HClusteringMethod { AVE, CEN, COM, MCQ, MED, SIN, WAR };

//! Mapping Two Dimensional Indexes into One Dimensional Index (Internal Function)
/*! \param N Size of the Square Matrix
 *  \param i Index 1
 *  \param j Index 2
 *
 *  Note that i must be less than j!
 *  \return Mapped One Dimensional Index
 *  \author Chibiao Chen
 */
inline int MapIndex(const int N, const int i, const int j)
{
	return N * i - ((i + 1) * (i + 2) >> 1) + j;
}

//! Building a Dissimilarity Matrix for Input Data (Internal Function)
/*! Measure the dissimilarity of each pair of input data and put the results into a
 *  dissimilarity matrix.
 *  \param X Input Data (One Dimensional)
 *  \param bG Genetic Data or Not
 *  \return The Dissimilarity Matrix (Represented by a Vector)
 *  \author Chibiao Chen
 */
template < typename T > DBL_VCTR_1D BuildDissimilarityMatrix(
	const vector< T >& X, 
	const bool bG = false)
{
	int N = X.size();
	int DML = static_cast<int>(0.5 * N * (N - 1));
	DBL_VCTR_1D dm(DML, 0.0);
	int i, j;
	if (bG)  // Genotype Data?
	{
		for (i = 0; i < N - 1; i++)
		{
			for (j = i + 1; j < N; j++)
			{
				if (X[i] == X[j])
				{
					dm[MapIndex(N, i, j)] = 0.0;
				}
				else
				{
					dm[MapIndex(N, i, j)] = 1.0;
				}
			}
		}
	}
	else
	{
		for (i = 0; i < N - 1; i++)
		{
			for (j = i + 1; j < N; j++)
			{
				dm[MapIndex(N, i, j)] = fabs(X[i] - X[j]);
			}
		}
	}

	return dm;
}

//! Building a Dissimilarity Matrix for Input Data (Internal Function)
/*! Measure the dissimilarity of each pair of input data and put the results into a
 *  dissimilarity matrix.
 *  \param X Input Data (Two Dimensional)
 *  \param bG Genetic Data or Not
 *  \return The Dissimilarity Matrix (Represented by a Vector)
 *  \author Chibiao Chen
 */
template < typename T > DBL_VCTR_1D BuildDissimilarityMatrix(
	const vector< vector< T > >& X, 
	const bool bG = false)
{
	int N = X.size();
	int TL = X[0].size();
	int L = TL >> 1;
	int DML = static_cast<int>(0.5 * N * (N - 1));
	DBL_VCTR_1D dm(DML, 0.0);
	int i, j, l;
	if (bG)  // Genotype Data? - Only Supporting Olivier's Format (GENECLUST, Diploid)
	{
		for (i = 0; i < N - 1; i++)
		{
			for (j = i + 1; j < N; j++)
			{
				int sim = 0;
				for (l = 0; l < TL; i++)
				{
					if (X[i][l] == X[j][l])
					{
						sim++;
					}
				}
				for (l = 0; l < L; l++)
				{
					if (X[i][2 * l] == X[j][2 * l + 1])
					{
						sim++;
					}
					if (X[i][2 * l + 1] == X[j][2 * l])
					{
						sim++;
					}
				}
				dm[MapIndex(N, i, j)] = 1.0 - static_cast<double>(sim) / (TL * 2.0);
			}
		}
	}
	else
	{
		for (i = 0; i < N - 1; i++)
		{
			for (j = i + 1; j < N; j++)
			{
				double dist = 0.0;
				for (l = 0; l < L; l++)
				{
					dist += (X[i][l] - X[j][l]) * (X[i][l] - X[j][l]);
				}
				dm[MapIndex(N, i, j)] = sqrt(dist);
			}
		}
	}

	return dm;
}

//! Building a Dissimilarity Matrix for Input Data (Internal Function)
/*! Measure the dissimilarity of each pair of input data and put the results into a
 *  dissimilarity matrix.
 *  \param X Input Data (Three Dimensional)
 *  \param bG Genetic Data or Not
 *  \return The Dissimilarity Matrix (Represented by a Vector)
 *  \author Chibiao Chen
 */
template < typename T > DBL_VCTR_1D BuildDissimilarityMatrix(
	const vector< vector< vector< T > > >& X, 
	const bool bG = false)
{
	int N = static_cast<int>(X.size());
	int A = static_cast<int>(X[0].size());
	int L = static_cast<int>(X[0][0].size());
	int DML = static_cast<int>(0.5 * N * (N - 1));
	DBL_VCTR_1D dm(DML, 0.0);
	int i, j, a, b, l;
	if (bG)  // Genotype Data? - Only Supporting Chibiao's Format
	{
		for (i = 0; i < N - 1; i++)
		{
			for (j = i + 1; j < N; j++)
			{
				int sim = 0;
				for (a = 0; a < A; a++)
				{
					for (b = 0; b < A; b++)
					{
						for (l = 0; l < L; l++)
						{
							if (X[i][a][l] == X[j][b][l])
							{
								sim++;
							}
						}
					}
				}
				dm[MapIndex(N, i, j)] = 1.0 - static_cast<double>(sim) / (A * A * L);
			}
		}
	}
	else
	{
		for (i = 0; i < N - 1; i++)
		{ 
			for (j = i + 1; j < N; j++)
			{
				double dist = 0.0;
				for (a = 0; a < A; a++)
				{
					for (l = 0; l < L; l++)
					{
						dist += (X[i][a][l] - X[j][a][l]) * (X[i][a][l] - X[j][a][l]);
					}
				}
				dm[MapIndex(N, i, j)] = sqrt(dist);
			}
		}
	}

	return dm;
}

//! The Hierarchical Clustering Engine (Internal Function)
/*! Do not call this internal function directly; call HClustering instead.
 *  \param N Number of Samples/Individuals in the Data
 *  \param dm Dissimilarity Matrix (Represented by a Vector)
 *  \param m Hierarchical Clustering Method
 *  \param ResultA Hierarchical Clustering Result Part A (Output)
 *  \param ResultB Hierarchical Clustering Result Part B (Output)
 *  \author Chibiao Chen
 */
void HClusteringEngine(
	const int N, 
	DBL_VCTR_1D& dm,
	HClusteringMethod m,
	INT_VCTR_1D& ResultA,
	INT_VCTR_1D& ResultB)
{
	INT_VCTR_1D Initializer(N - 1, 0);
	if (ResultA.size() == 0)
	{
		ResultA = Initializer;
	}
	if (ResultB.size() == 0)
	{
		ResultB = Initializer;
	}

	const double INF = 1.0E+300;

	INT_VCTR_1D NoM(N, 1);				// Number of Members in Each Cluster
	BOOL_VCTR_1D Clustered(N, false);	// 
	int NoC = 0;						// Number of Clustered
	INT_VCTR_1D NN(N - 1, 0);			// Nearest Neighbors
	DBL_VCTR_1D DisNN(N - 1, INF);		// Dissimilarities of Nearest Neighbors

	for (NoC = 0; NoC < N - 1; NoC++)
	{
		int i, j, k;

		double dMin;

		int IMin, JMin;
		for (i = 0; i < N - 1; i++)
		{
			if (Clustered[i])
			{
				continue;
			}
			dMin = INF;
			for (j = i + 1; j < N; j++)
			{
				if (Clustered[j])
				{
					continue;
				}
				int Index = MapIndex(N, i, j); 
				if (dm[Index] < dMin)
				{
					dMin = dm[Index];
					JMin = j;
				}
			}
			NN[i] = JMin;
			DisNN[i] = dMin;
		}
		dMin = INF;
		for (i = 0; i < N - 1; i++)
		{
			if (Clustered[i])
			{
				continue;
			}
			if (DisNN[i] < dMin)
			{
				dMin = DisNN[i];
				IMin = i;
				JMin = NN[i];
			}
		}
		int IMerge, JMerge;
		IMerge = IMin < JMin ? IMin : JMin;
		JMerge = IMin > JMin ? IMin : JMin;
		ResultA[NoC] = IMerge;
		ResultB[NoC] = JMerge;
		Clustered[JMerge] = true;

		dMin = INF;
		int IndexIJ, IndexIK, IndexJK;
		for (k = 0; k < N; k++)
		{
			if (Clustered[k])
			{
				continue;
			}
			if (k == IMerge)
			{
				continue;
			}
			IndexIJ = MapIndex(N, IMerge, JMerge);
			if (IMerge < k)
			{
				IndexIK = MapIndex(N, IMerge, k);
			}
			else
			{
				IndexIK = MapIndex(N, k, IMerge);
			}
			if (JMerge < k)
			{
				IndexJK = MapIndex(N, JMerge, k);
			}
			else
			{
				IndexJK = MapIndex(N, k, JMerge);
			}
			int NoMIJ = NoM[IMerge] + NoM[JMerge];
			switch(m) 
			{
				case AVE:
					dm[IndexIK] = (dm[IndexIK] * NoM[IMerge] + dm[IndexJK] * NoM[JMerge]) 
						/ NoMIJ;
					break;

				case CEN:
					dm[IndexIK] = (dm[IndexIK] * NoM[IMerge] + dm[IndexJK] * NoM[JMerge] -
						dm[IndexIJ] * NoM[IMerge] * NoM[JMerge] / NoMIJ) / NoMIJ;
					break;

				case COM:
					dm[IndexIK] = max(dm[IndexIK], dm[IndexJK]);
					break;

				case MCQ:
					dm[IndexIK] = 0.5 * (dm[IndexIK] + dm[IndexJK]);
					break;

				case MED:
					dm[IndexIK] = 0.5 * (dm[IndexIK] + dm[IndexJK]) - 0.25 * dm[IndexIJ];
					break;

				case SIN:
					dm[IndexIK] = min(dm[IndexIK], dm[IndexJK]);
					break;

				case WAR:
					dm[IndexIK] = (dm[IndexIK] * (NoM[IMerge] + NoM[k]) + 
						dm[IndexJK] * (NoM[JMerge] + NoM[k]) - dm[IndexIJ] * NoM[k]) / 
						(NoMIJ + NoM[k]);
					break;

				default:
					cerr << "Unsupported Clustering Method - HClusteringEngine" << endl;
					exit(EXIT_FAILURE);
			}
		}
		NoM[IMerge] += NoM[JMerge];
	}
}

//! Labeling Samples/Individuals into Clusters (Internal Function)
/*! Do not call this internal function directly; call HClustering instead.
 *  \param ResultA Hierarchical Clustering Result Part A
 *  \param ResultB Hierarchical Clustering Result Part B
 *  \param K Number of Clusters
 *  \return Cluster Labels of Samples/Individuals
 *  \author Chibiao Chen
 */
INT_VCTR_1D LabelSamples(const INT_VCTR_1D& ResultA, const INT_VCTR_1D& ResultB, int K)
{
	int i, j;
	int N = static_cast<int>(ResultA.size()) + 1;
	BOOL_VCTR_1D Merged(N, false);
	INT_VCTR_2D Members;
	for (i = 0; i < N; i++)
	{
		INT_VCTR_1D SingleElementVector(1, i);
		Members.push_back(SingleElementVector);
	}
	for (i = 0; i < N - K; i++)
	{
		for (j = 0; j < static_cast<int>(Members[ResultB[i]].size()); j++)
		{
			Members[ResultA[i]].push_back(Members[ResultB[i]][j]);
		}
		Merged[ResultB[i]] = true;
	}
	INT_VCTR_1D Labels(N, 0);
	int k = 0;
	for (i = 0; i < N; i++)
	{
		if (!Merged[i])
		{
			for (j = 0; j < static_cast<int>(Members[i].size()); j++)
			{
				Labels[Members[i][j]] = k;
			}
			k++;
		}
	}

	return Labels;
}

//! Hierarchical Clustering
/*! \param X Data to Be Clustered (One Dimensional)
 *  \param m Hierarchical Clustering Method
 *  \param NoC Number of Clusters
 *  \param bG Genetic Data or Not
 *  \return The Hierarchical Clustering Results
 *  \author Chibiao Chen
 */
template < typename T > INT_VCTR_1D HClustering(
	const vector< T >& X,		// Data to Be Clustered 
	const HClusteringMethod m,
	const int NoC,				// Number of Clusters
	const bool bG = false)
{
	int N = X.size();  // Number of Samples
	if (N == 0)
	{
		cerr << "Empty Data - HClustering" << endl;
		exit(EXIT_FAILURE);
	}
	DBL_VCTR_1D dm = BuildDissimilarityMatrix(X, bG);
	INT_VCTR_1D ResultA(N - 1, 0);
	INT_VCTR_1D ResultB(N - 1, 0);
	HClusteringEngine(N, dm, m, ResultA, ResultB);
	return LabelSamples(ResultA, ResultB, NoC);
}

//! Hierarchical Clustering
/*! \param X Data to Be Clustered (Two Dimensional)
 *  \param m Hierarchical Clustering Method
 *  \param NoC Number of Clusters
 *  \param bG Genetic Data or Not
 *  \return The Hierarchical Clustering Results
 *  \author Chibiao Chen
 */
template < typename T > INT_VCTR_1D HClustering(
	const vector< vector< T > >& X,	// Data to Be Clustered 
	const HClusteringMethod m,
	const int NoC,					// Number of Clusters
	const bool bG = false)
{
	int N = X.size();  // Number of Samples
	if (N == 0)
	{
		cerr << "Empty Data - HClustering" << endl;
		exit(EXIT_FAILURE);
	}
	DBL_VCTR_1D dm = BuildDissimilarityMatrix(X, bG);
	INT_VCTR_1D ResultA(N - 1, 0);
	INT_VCTR_1D ResultB(N - 1, 0);
	HClusteringEngine(N, dm, m, ResultA, ResultB);
	return LabelSamples(ResultA, ResultB, NoC);
}

//! Hierarchical Clustering
/*! \param X Data to Be Clustered (Three Dimensional)
 *  \param m Hierarchical Clustering Method
 *  \param NoC Number of Clusters
 *  \param bG Genetic Data or Not
 *  \return The Hierarchical Clustering Results
 *  \author Chibiao Chen
 */
template < typename T > INT_VCTR_1D HClustering(
	const vector< vector< vector< T > > >& X,	// Data to Be Clustered 
	const HClusteringMethod m,
	const int NoC,								// Number of Clusters
	const bool bG = false)
{
	int N = static_cast<int>(X.size());  // Number of Samples
	if (N == 0)
	{
		cerr << "Empty Data - HClustering" << endl;
		exit(EXIT_FAILURE);
	}
	DBL_VCTR_1D dm = BuildDissimilarityMatrix(X, bG);
	INT_VCTR_1D ResultA(N - 1, 0);
	INT_VCTR_1D ResultB(N - 1, 0);
	HClusteringEngine(N, dm, m, ResultA, ResultB);
	return LabelSamples(ResultA, ResultB, NoC);
}

#endif//__HIERARCHICAL_CLUSTERING__