#include <cstdlib>
#include <fstream>
#include <iostream>

#include "IOUtilities.h"
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sort.h>

using namespace std;

void computeSymmMatrixStat(double& stat, const DBL_VCTR_2D& Matrix, const string option)
{
	if (Matrix.size() == 0)
	{
		cerr << "Uninitialized Matrix - computeSymmMatrixStat" << endl;
		exit(EXIT_FAILURE);
	}
	
	int N = static_cast<int>(Matrix.size());
	int i,j;
	int cpt = 0;
	int size = (N*(N+1)/2); //size of upper triangle
// 	gsl_vector* gslVect = gsl_vector_alloc(size);
	double vect[size];
	for (i = 0; i < N; i++)
	{
		for (j = i; j < N; j++)
		{
// 			gsl_vector_set(gslVect,cpt,Matrix[i][j]);
			vect[cpt] = Matrix[i][j];
			cpt +=1;
		}
	}
	
	bool goodOpt = false;
	if (option == "m") 
	{
		stat = gsl_stats_mean(vect,1,size);
		goodOpt = true;
	}
	if (option == "d")
	{
		gsl_sort(vect,1,size);
		stat = gsl_stats_median_from_sorted_data(vect,1,size);
		goodOpt = true;
	}
	if (option == "x")
	{
		stat = gsl_stats_max(vect,1,size);
		goodOpt = true;
	}
	
	if (!goodOpt)
	{
		cerr << "Unknown option for scale parameter - computeSymmMatrixStat" << endl;
		exit(EXIT_FAILURE);
	}
// 	gsl_vector_free(gslVect);	
	
}


bool SaveData(const DBL_VCTR_2D& C, const INT_VCTR_3D& X, const string& File)
{
	if (C.size() == 0)
	{
		cerr << "Uninitialized C - SaveData" << endl;
		exit(EXIT_FAILURE);
	}

	if (X.size() == 0)
	{
		cerr << "Uninitialized X - SaveData" << endl;
		exit(EXIT_FAILURE);
	}

	if (C.size() < X.size())
	{
		cerr << "Mismatched C and X - SaveData" << endl;
		exit(EXIT_FAILURE);
	}

	int i, a, l;

	ofstream ofs(File.c_str());
	if (!ofs.good())
	{
		cerr << "Cannot open file " << File << " - SaveData" << endl;
		exit(EXIT_FAILURE);
	}

	int M = static_cast<int>(C.size());
	int N = static_cast<int>(X.size());
	int A = static_cast<int>(X[0].size());
	int L = static_cast<int>(X[0][0].size());
	for (i = 0; i < M; i++)
	{
		for (a = 0; a < A; a++)
		{
			ofs << C[i][0] << "\t" << C[i][1] << "\t";
			if (i < N)
			{
				for (l = 0; l < L; l++)
				{
					ofs << X[i][a][l] << "\t";
				}
			}
			else
			{
				for (l = 0; l < L; l++)
				{
					ofs << 0 << "\t";
				}
			}
			ofs << endl;
		}
	}

	ofs.close();

	return true;
}

bool LoadData(DBL_VCTR_2D& C, INT_VCTR_3D& X, int ER, int EC, const string& File, const bool& specialFormat)
{
	if (C.size() == 0)
	{
		cerr << "Uninitialized C - LoadData" << endl;
		exit(EXIT_FAILURE);
	}

	if (X.size() == 0)
	{
		cerr << "Uninitialized X - LoadData" << endl;
		exit(EXIT_FAILURE);
	}

	if (C.size() < X.size())
	{
		cerr << "Mismatched C and X - LoadData" << endl;
		exit(EXIT_FAILURE);
	}

	int i, a, l;

	ifstream ifs(File.c_str());
	if (!ifs.good())
	{
		cerr << "Cannot open file " << File << " - LoadData" << endl;
		exit(EXIT_FAILURE);
	}

	int M = static_cast<int>(C.size());
	int N = static_cast<int>(X.size());
	int A = static_cast<int>(X[0].size());
	int L = static_cast<int>(X[0][0].size());

	// Load and discard the extra rows.
	char extraRow[10240];
	for (int er = 0; er < ER; er++)
	{
		ifs.getline(extraRow, 10240);
	}
	if (!specialFormat)
	{
		for (i = 0; i < M; i++)
		{
			for (a = 0; a < A; a++)
			{
				// Load and discard the extra columns.
				string discarded;
				for (int ec = 0; ec < EC; ec++)
				{
					ifs >> discarded;
				}
				//
				ifs >> C[i][0] >> C[i][1];
				if (i < N)
				{
					for (l = 0; l < L; l++)
					{
						ifs >> X[i][a][l];
					}
				}
				else
				{
					for (l = 0; l < L; l++)
					{
						ifs >> discarded;
					}
				}
			}
		}
	}
	
	if (specialFormat)
	{
		for (i = 0; i < M; i++)
		{
			// Load and discard the extra columns.
			string discarded;
			for (int ec = 0; ec < EC; ec++)
			{
				ifs >> discarded;
			}
			ifs >> C[i][0] >> C[i][1];
			if (i < N)
			{
				for (a = 0; a < A; a++)
				{
					for (l = 0; l < L; l++)
					{
						ifs >> X[i][a][l];
					}
				}
			}
			else
			{
				for (a = 0; a < A; a++)
				{
					for (l = 0; l < L; l++)
					{
						ifs >> discarded;
					}
				}
			}
		}
	}
	
	ifs.close();

	return true;
}

bool SaveCF(const INT_VCTR_1D& z, const DBL_VCTR_3D& p, const string& File)
{
	if (z.size() == 0)
	{
		cerr << "Uninitialized z - SaveCF" << endl;
		return false;
	}

	if (p.size() == 0)
	{
		cerr << "Uninitialized p - SaveCF" << endl;
		return false;
	}

	int i, k, l, j;

	ofstream ofs(File.c_str());
	if (!ofs.good())
	{
		cerr << "Cannot open file " << File << " - SaveCF" << endl;
		return false;
	}

	// K, L, J
	int K = static_cast<int>(p.size());
	int L = static_cast<int>(p[0].size());
	INT_VCTR_1D J(L, 0);
	for (l = 0; l < L; l++)
	{
		J[l] = static_cast<int>(p[0][l].size());
	}
	ofs << K << ' ';
	ofs << L << ' ';
	for (l = 0; l < L; l++)
	{
		ofs << J[l] << ' ';
	}

	// N
	int N = static_cast<int>(z.size());
	ofs << N << ' ';

	// p
	for (k = 0; k < K; k++)
	{
		for (l = 0; l < L; l++)
		{
			for (j = 0; j < J[l]; j++)
			{
				ofs << p[k][l][j] << ' ';
			}
		}
	}

	// z
	for (i = 0; i < N; i++)
	{
		ofs << z[i] << ' ';
	}

	ofs.close();

	return true;
}

bool LoadCF(INT_VCTR_1D& z, DBL_VCTR_3D& p, const string& File)
{
	if (z.size() == 0)
	{
		cerr << "Uninitialized z - LoadCF" << endl;
		return false;
	}

	if (p.size() == 0)
	{
		cerr << "Uninitialized p - LoadCF" << endl;
		return false;
	}

	int i, k, l, j;

	// K, L, J
	int K = static_cast<int>(p.size());
	int L = static_cast<int>(p[0].size());
	INT_VCTR_1D J(L, 0);
	for (l = 0; l < L; l++)
	{
		J[l] = static_cast<int>(p[0][l].size());
	}

	// N
	int N = static_cast<int>(z.size());

	ifstream ifs(File.c_str());
	if (!ifs.good())
	{
		cerr << "Cannot open file " << File << " - LoadCF" << endl;
		return false;
	}

	// KF, LF, JF
	int KF;
	int LF;
	INT_VCTR_1D JF;
	ifs >> KF;
	ifs >> LF;
	for (l = 0; l < LF; l++)
	{
		JF.push_back(0);
		ifs >> JF[l];
	}

	// NF
	int NF;
	ifs >> NF;
	if (KF != K || LF != L || JF != J || NF != N)
	{
		ifs.close();

		cerr << "CF file conflicts with command-line options" << " - LoadCF" << endl;
		return false;
	}

	for (k = 0; k < KF; k++)
	{
		for (l = 0; l < LF; l++)
		{
			for (j = 0; j < JF[l]; j++)
			{
				ifs >> p[k][l][j];
			}
		}
	}

	for (i = 0; i < NF; i++)
	{
		ifs >> z[i];
	}

	ifs.close();

	return true;
}

bool SaveML(const double& psi, const double& minDIC,const string& File)
{
	ofstream ofs(File.c_str());
	if (!ofs.good())
	{
		cerr << "Cannot open file " << File << " - SaveML" << endl;
		return false;
	}
	ofs << minDIC << ' ';
	ofs << psi << ' ';
	ofs << "0" << ' '; 	//admixture = false
	ofs.close();
	return true;
}

bool LoadML(double& psi, double& minDIC ,const string& File,bool quiet)
{
	ifstream ifs(File.c_str());
	if (!ifs.good())
	{
		if (!quiet)
		{
			cerr << "Cannot open file " << File << " - LoadML" << endl;
		}
		return false;
	}
	
	ifs >> minDIC ;
	ifs >> psi;
	ifs.close();
	return true;
}
