#include <chartdir.h>
#include <iostream>
#include <set>

#include "PlotUtilities.h"
#include "VoronoiGeometry.h"

using namespace std;

#define NUM_COL 32
#define NUM_SYM 32

int COLORS[NUM_COL] = {
	0xFF3333, 0x33FF33, 0x6666FF, 0xFFFF00, 0xFF66FF, 0x99FFFF, 0xFFCC33, 0xCCCCCC, 
	0xCC9999, 0x339966, 0x999900, 0xCC3300, 0x669999, 0x993333, 0x006600, 0x990099, 
	0xFF9966, 0x99FF99, 0x9999FF, 0xCC6600, 0x33CC33, 0xCC99FF, 0xFF6666, 0x99CC66, 
	0x009999, 0xCC3333, 0x9933FF, 0xFF0000, 0x0000FF, 0x00FF00, 0xFFCC99, 0x999999 
};

// Draw a bar chart to show the assignment probabilities.
int PlotAP(const char* FileName, const DBL_VCTR_2D& ap)
{
	int i, k;

	int N = static_cast<int>(ap.size());
	int K = static_cast<int>(ap[0].size());

	DBL_VCTR_2D chartData;
	InitVctr2D(chartData, K, N, 0.0);
	for (i = 0; i < N; i++)
	{
		for (k = 0; k < K; k++)
		{
			chartData[k][i] = ap[i][k];
		}
	}

	DBL_VCTR_1D chartLabels(N, 0.0);
	for (i = 0; i < N; i++)
	{
		// Make sure the first index is 1 not 0.
		chartLabels[i] = i + 1;
	}

	int barWidth = 30;
	XYChart* c = new XYChart(barWidth * N + 100, 600, 0xFFFFCC, 0, 1);
	c->setPlotArea(50, 50, barWidth * N, 500, 0xFFFFFF);
	c->setClipping();
	c->addTitle("Assignment Probabilities", "timesbi.ttf", 14, 0xFFFFFF
		)->setBackground(0x804020);
	c->xAxis()->setLabels(DoubleArray(&chartLabels[0], N), "{value|0}");
	BarLayer *layer = c->addBarLayer(Chart::Percentage);
	c->yAxis()->setLabelFormat("{={value}*0.01}");
	c->yAxis()->setTickDensity(50);

	for (k = 0; k < K; k++)
	{
		layer->addDataSet(DoubleArray(&chartData[k][0], N));
	}
	layer->setBorderColor(Chart::Transparent);
	bool bMade = c->makeChart(FileName);
	if (!bMade)
	{
		cerr << "Unable to make chart - PlotAP" << endl;
	}
	delete c;

	return 0;
}

// Draw a line chart to show the log-likelihood history.
int PlotLH(const char* FileName, const DBL_VCTR_1D& lh)
{
	int i;

	DBL_VCTR_1D lhCopy = lh;
	int NoDP = static_cast<int>(lhCopy.size());  // Number of Data Points

	const double NEG_INF = -1.0E+300;
	for (i = 0; i < NoDP; i++)
	{
		if (lhCopy[i] < NEG_INF)
		{
			cerr << "Data Out Of Range, No Chart Plotted - PlotLH" << endl;
			return -1;
		}
	}

	DBL_VCTR_1D chartLabels(NoDP, 0.0);
	for (i = 0; i < NoDP; i++)
	{
		// Make sure the first index is 1 not 0.
		chartLabels[i] = i + 1;
	}

	XYChart* c = new XYChart(900, 600, 0xFFFFCC, 0, 1);
	c->setPlotArea(50, 50, 800, 500, 0xFFFFFF);
	c->setClipping();
	c->addTitle("Log-Likelihood History", "timesbi.ttf", 14, 0xFFFFFF
		)->setBackground(0x804020);
	c->addLineLayer(DoubleArray(&lhCopy[0], NoDP));
	c->xAxis()->setLabels(DoubleArray(&chartLabels[0], NoDP), "{value|0}");
	int nStep = NoDP / 10;
	if (nStep > 0)
	{
		c->xAxis()->setLabelStep(nStep);
	}
	bool bMade = c->makeChart(FileName);
	if (!bMade)
	{
		cerr << "Unable to make chart - PlotLH" << endl;
	}
	delete c;

	return 0;
}

// Draw a scatter chart to show the hard (spatial) clustering result.
int PlotHC(const char* FileName, 
		   const vector< SPoint >& vptX, 
		   const vector< double >& vdDataScale, 
		   const INT_VCTR_1D& hc, 
		   const int NumOfC, 
		   bool bDrawLabel /* = false */)
{
	if (vptX.size() == 0)
	{
		cerr << "No Voronoi Points - PlotHC" << endl;
		return -1;
	}
	if (vdDataScale.size() != 4)
	{
		cerr << "Wrong Data Scale - PlotHC" << endl;
		return -1;	
	}
	if (hc.size() == 0)
	{
		cerr << "No Hard Clustering Results - PlotHC" << endl;
		return -1;
	}
	if (hc.size() > vptX.size())
	{
		cerr << "Wrong Hard Clustering Results - PlotHC" << endl;
		return -1;
	}

	int ScatterSymbols[NUM_SYM] = {
		Chart::SquareSymbol, 
		Chart::DiamondSymbol, 
		Chart::TriangleSymbol, 
		Chart::RightTriangleSymbol, 
		Chart::LeftTriangleSymbol, 
		Chart::InvertedTriangleSymbol, 
		Chart::CircleSymbol, 
		Chart::StarShape(3), 
		Chart::StarShape(4), 
		Chart::StarShape(5), 
		Chart::StarShape(6), 
		Chart::StarShape(7), 
		Chart::StarShape(8), 
		Chart::StarShape(9), 
		Chart::StarShape(10), 
		Chart::PolygonShape(5), 
		Chart::Polygon2Shape(5), 
		Chart::PolygonShape(6), 
		Chart::Polygon2Shape(6), 
		Chart::PolygonShape(7), 
		Chart::PolygonShape(8), 
		Chart::CrossShape(0.1), 
		Chart::CrossShape(0.2), 
		Chart::CrossShape(0.3), 
		Chart::CrossShape(0.4), 
		Chart::CrossShape(0.5), 
		Chart::CrossShape(0.6), 
		Chart::CrossShape(0.7), 
		Chart::Cross2Shape(0.1), 
		Chart::Cross2Shape(0.2), 
		Chart::Cross2Shape(0.3), 
		Chart::Cross2Shape(0.4) 
//		Chart::Cross2Shape(0.5), 
//		Chart::Cross2Shape(0.6), 
//		Chart::Cross2Shape(0.7)
	};

	int i, j;
	// int N = static_cast<int>(vptX.size());
	int N = static_cast<int>(hc.size());
	set< int > ClusterSet;
	int NoC;
	if (NumOfC < 1)
	{
		for (i = 0; i < N; i++)
		{
			ClusterSet.insert(hc[i]);
		}
		NoC = static_cast<int>(ClusterSet.size());
	}
	else
	{
		for (i = 0; i < NumOfC; i++)
		{
			ClusterSet.insert(i);
		}
		NoC = NumOfC;
	}
	vector< int > ClusterVec;
	set< int >::const_iterator csiB = ClusterSet.begin();
	set< int >::const_iterator csiE = ClusterSet.end();
	set< int >::const_iterator csi;
	for (csi = csiB; csi != csiE; ++csi)
	{
		ClusterVec.push_back(*csi);
	}

	double dW = vdDataScale[1] - vdDataScale[0];
	double dH = vdDataScale[3] - vdDataScale[2];
	int nW;
	int nH;
	int nShortSide = 500;
	if (dW > dH)
	{
		nH = nShortSide;
		nW = int(nH * dW / dH + 0.5);
	}
	else
	{
		nW = nShortSide;
		nH = int(nW * dH / dW + 0.5);
	}
	int nXE = 50;
	int nYE = 50;
	XYChart* c = new XYChart(nW + 2 * nXE, nH + 2 * nYE, 0xFFFFCC, 0, 1);
	c->addTitle("Membership of Individuals", "timesbi.ttf", 14, 0xFFFFFF
		)->setBackground(0x804020);
	c->setPlotArea(nXE, nYE, nW, nH, 0xFFFFFF, -1, -1, Transparent, Transparent);
	c->setClipping();
	c->xAxis()->setDateScale(vdDataScale[0], vdDataScale[1]);
	c->xAxis()->setAutoScale(0.0, 0.0, 0.0);
	c->xAxis()->setTickDensity(nShortSide / 10);
	c->xAxis()->setColors(LineColor, Transparent, Transparent, Transparent);
	c->yAxis()->setDateScale(vdDataScale[2], vdDataScale[3]);
	c->yAxis()->setAutoScale(0.0, 0.0, 0.0);
	c->yAxis()->setTickDensity(nShortSide / 10);
	c->yAxis()->setColors(LineColor, Transparent, Transparent, Transparent);
	vector< double > vdXC;
	vector< double > vdYC;
	vector< double > vdIndex;
	ScatterLayer* pslPoints;
	for (i = 0; i < NoC; i++)
	{
		vdXC.clear();
		vdYC.clear();
		vdIndex.clear();
		for (j = 0; j < N; j++)
		{
			if (hc[j] == ClusterVec[i])
			{
				vdXC.push_back(vptX[j].m_psXY.m_dX);
				vdYC.push_back(vptX[j].m_psXY.m_dY);
				// Make sure the first index is 1 not 0.
				vdIndex.push_back(vptX[j].m_nIndex + 1);
			}
		}
		pslPoints = c->addScatterLayer(
			DoubleArray(&vdXC[0], (int) vdXC.size()),
			DoubleArray(&vdYC[0], (int) vdYC.size()),
			"Individuals", ScatterSymbols[ClusterVec[i]], 9);
		if (bDrawLabel)
		{
			pslPoints->addExtraField(DoubleArray(&vdIndex[0], (int) vdIndex.size()));
			pslPoints->setDataLabelFormat("{field0}");
			TextBox* ptbLabel = pslPoints->setDataLabelStyle();
			ptbLabel->setBackground(Transparent);
			ptbLabel->setAlignment(Chart::Left);
			ptbLabel->setPos(1, 0);
		}
	}
	c->makeChart(FileName);
	delete c;

	return 0;
}

// Draw a chart to show the hard (spatial) clustering result.
int PlotHC(const char* FileName, 
		   const vector< SPoint >& vptX,
		   const vector< bool >& vbKeep, 
		   const vector< double >& vdDataMinMax, 
		   vector< CEdge >& veVoronoi, 
		   const INT_VCTR_1D& hc)
{
	if (vptX.size() == 0)
	{
		cerr << "No Voronoi Points - PlotHC" << endl;
		return -1;
	}
	if (vbKeep.size() == 0)
	{
		cerr << "No Keep Flags - PlotHC" << endl;
		return -1;
	}
	if (vdDataMinMax.size() != 4)
	{
		cerr << "Wrong Data Min Max - PlotHC" << endl;
		return -1;	
	}
	if (veVoronoi.size() == 0)
	{
		cerr << "No Voronoi Edges - PlotHC" << endl;
		return -1;
	}
	if (hc.size() == 0)
	{
		cerr << "No Hard Clustering Results - PlotHC" << endl;
		return -1;
	}
	if (hc.size() > vptX.size())
	{
		cerr << "Wrong Hard Clustering Results - PlotHC" << endl;
		return -1;
	}

	CVoronoiGeometry vgThis(vdDataMinMax[0], vdDataMinMax[1], 
		vdDataMinMax[2], vdDataMinMax[3]);
	double dXMaxLmt = vgThis.GetXMaxLmt();
	double dXMinLmt = vgThis.GetXMinLmt();
	double dYMaxLmt = vgThis.GetYMaxLmt();
	double dYMinLmt = vgThis.GetYMinLmt();
	double dW = dXMaxLmt - dXMinLmt;
	double dH = dYMaxLmt - dYMinLmt;
	double dShortSide;
	int nW;
	int nH;
	int nShortSide = 500;
	if (dW > dH)
	{
		dShortSide = dH;
		nH = nShortSide;
		nW = int(nH * dW / dH + 0.5);
	}
	else
	{
		dShortSide = dW;
		nW = nShortSide;
		nH = int(nW * dH / dW + 0.5);
	}
	int nXE = 50;
	int nYE = 50;
	XYChart* c = new XYChart(nW + 2 * nXE, nH + 2 * nYE, 0xFFFFCC, 0, 1);
	c->addTitle("Membership of Individuals", "timesbi.ttf", 14, 0xFFFFFF
		)->setBackground(0x804020);
	int bgColor = 0xFFFFFF;
	c->setPlotArea(nXE, nYE, nW, nH, bgColor, -1, -1, Transparent, Transparent);
	c->setClipping();
	c->xAxis()->setDateScale(dXMinLmt, dXMaxLmt);
	c->xAxis()->setAutoScale(0.0, 0.0, 0.0);
	c->xAxis()->setTickDensity(nShortSide / 10);
	c->xAxis()->setColors(LineColor, Transparent, Transparent, Transparent);
	c->yAxis()->setDateScale(dYMinLmt, dYMaxLmt);
	c->yAxis()->setAutoScale(0.0, 0.0, 0.0);
	c->yAxis()->setTickDensity(nShortSide / 10);
	c->yAxis()->setColors(LineColor, Transparent, Transparent, Transparent);
	SPosition psOne;
	SPosition psTwo;
	double daX[2];
	double daY[2];
	int lnColor = 0x000000;
	int i;
	for (i = 0; i < (int) veVoronoi.size(); i++)
	{
		SEdge eThis;
		CVoronoi::CopyEdge(&veVoronoi[i], &eThis);
		vgThis.ClipEdge(&eThis, psOne, psTwo);
		daX[0] = psOne.m_dX;
		daY[0] = psOne.m_dY;
		daX[1] = psTwo.m_dX;
		daY[1] = psTwo.m_dY;
		if (daX[0] != INF && daY[0] != INF && daX[1] != INF && daY[1] != INF)
		{
			LineLayer* pll = c->addLineLayer(DoubleArray(daY, 2), lnColor);
			pll->setXData(DoubleArray(daX, 2));
		}
	}
	DrawArea* da = c->makeChart();
	da->vFlip();
	double dFactor = nShortSide / dShortSide;
	int NoTP = static_cast<int>(vptX.size());  // Number of Total Points
	vector< int > vnXC(NoTP, 0);
	vector< int > vnYC(NoTP, 0);
	for (i = 0; i < NoTP; i++)
	{
		vnXC[i] = int(nXE + (vptX[i].m_psXY.m_dX - dXMinLmt) * dFactor + 0.5);
		vnYC[i] = int(nYE + (vptX[i].m_psXY.m_dY - dYMinLmt) * dFactor + 0.5);
	}
	int NoP = static_cast<int>(hc.size());  // Number of Data Points
	for (i = 0; i < NoP; i++)  // Initial Filling
	{
		if (vbKeep[i])
		{
			if (da->getPixel(vnXC[i], vnYC[i]) != bgColor)
			{
				continue;
			}
			da->fill(vnXC[i], vnYC[i], COLORS[hc[i]]);
		}
	}
	vector< int > vnD(NoTP, 0x7FFFFFFF);  // Second Filling; Slow But Reliable
	int j, k;
	for (i = nXE; i < nW + nXE; i++)
	{
		for (j = nYE; j < nH + nYE; j++)
		{
			if (da->getPixel(i, j) == bgColor)
			{
				for (k = 0; k < NoTP; k++)
				{
					if (vbKeep[k])
					{
						vnD[k] = (i - vnXC[k]) * (i - vnXC[k]) + 
							(j - vnYC[k]) * (j - vnYC[k]);
					}
				}
				int nMin = 0x7FFFFFFF;
				int kMin = 0;
				for (k = 0; k < NoTP; k++)
				{
					if (vbKeep[k])
					{
						if (vnD[k] < nMin)
						{
							nMin = vnD[k];
							kMin = k;
						}
					}
				}
				if (kMin < NoP)
				{
					da->fill(i, j, COLORS[hc[kMin]]);
				}
			}
		}
	}
	for (i = NoP; i < NoTP; i++)
	{
		if (vbKeep[i])
		{
			da->fill(vnXC[i], vnYC[i], bgColor);
		}
	}
	for (i = 0; i < NoP; i++)
	{
		if (vbKeep[i])
		{
			da->circle(vnXC[i], vnYC[i], 2, 2, 0x000000, 0x000000);
		}
	}
	for (i = NoP; i < NoTP; i++)
	{
		if (vbKeep[i])
		{
			da->circle(vnXC[i], vnYC[i], 2, 2, 0x000000, 0xFFFFFF);
		}
	}
	da->vFlip();
	da->out(FileName);
	delete c;

	return 0;
}
