/*! \file tessm.cpp
 *  \brief The TESS MCMC Algorithms
 */

#include <algorithm>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <sstream>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>

#include <iostream>
#include <set>
#include <string>

#include "IOUtilities.h"
#include "PlotUtilities.h"
#include "RNGUtilities.h"
#include "VectorUtilities.h"
#include "Voronoi.h"
#include "HierarchicalClustering.h"

//! The Negative Limit of Logarithm (double)
#define N_LIMIT (-707.0)

//! The Positive Limit of Logarithm (double)
#define P_LIMIT (+707.0)

//! Sweep Parameters
class CSweepParams
{
public:
	INT_VCTR_1D z;	//!< Cluster Label of Individuals
	DBL_VCTR_3D p;	//!< Allele Frequencies
};

using namespace std;

//! Showing the Usage of the Program
/*! The compulsory/required options are indicated by upper case characters; the optional
 *  options are indicated by lower case characters.
 *  \author Chibiao Chen
 */
void ShowUsage()
{
	cout <<	"Typical Usage: tessm -FF -NN -AA -LL -KK -PP -DD -SS -BB" << endl;
	cout << "Options can be specified in any order." << endl;
	cout << endl;
	cout << "Required Options:" << endl;
	cout << "-FF: F = File Name of Input Data File" << endl;
	cout << "     For file name contains spaces, input it as \"F\"." << endl;
	cout << "-NN: N = Number of Individuals" << endl;
	cout << "-AA: A = Ploidy (1 = Haploid, 2 = Diploid, ...)" << endl;
	cout << "-LL: L = Number of Loci" << endl;
	cout << "-KK: K = Number of Clusters" << endl;
	cout << "-PP: P = Interaction Parameter of HMRF" << endl;
	cout << "-DD: D = Parameter of Dirichlet Allele Frequency Model" << endl;
	cout << "-SS: S = Total Number of Sweeps of MCMC" << endl;
	cout << "-BB: B = Burn In Number of Sweeps of MCMC" << endl;
	cout << endl;
	cout << "Optional Options:" << endl;
	cout << "-rr: r = Number of Extra Rows in Data File, Default: 0" << endl;
	cout << "-cc: c = Number of Extra Columns in Data File, Default: 0" << endl;
	cout << "-dd: d = Number of Dummy Individuals, Default: 0" << endl;
	cout << "-ii: i = Folder Name of Input Data File, Default: Current Folder" << endl;
	cout << "     For folder name contains spaces, input it as \"i\"." << endl;
	cout << "-jj: j = Init the run with Neighbor Joining tree" << endl;
	cout << "     j = y, Yes" << endl;
	cout << "     j = n, No (Default)" << endl;
	cout << "-oo: o = Folder Name of Output Result Files, Default: Current Folder"
		<< endl;
	cout << "     For folder name contains spaces, input it as \"o\"." << endl;
	cout << "-pp: p = Configuration File for Continue from Previous Run, Default: NULL"
		<< endl;
	cout << "     For file name contains spaces, input it as \"p\"." << endl;
	cout << "-spsp: sp = SPecial data format: one individual = one row" << endl;
	cout << "       sp = y, Yes" << endl;
	cout << "       sp = n, No (Default)" << endl;
	cout << "-gg: g = Name of file containing Geographic distances between individuals (optional) " << endl;
	cout << "         For file name contains spaces, input it as \"g\"." << endl;
	cout << "-scsc: sc = scale parameter for geographic distances. User predefined value or enter one. Ignored if no geographic distance file is given." << endl;
	cout << "       sc = m, Mean distance between individuals (default)" << endl;
	cout << "       sc = d, meDian distance between individuals" << endl; 
	cout << "       sc = x, maX distance between individuals" << endl;
	cout << "       sc = value of the scale parameter (e.g. 10.4)" << endl;
//	cout << "-vv: v = Configuration File of Neighborhood System, Default: NULL" << endl;
//	cout << "     For file name contains spaces, input it as \"v\"." << endl;
}

//! Translating the XY Coordinates
/*! Translate the XY coordinates to all positive values to circumvent the quirks of
 *  ChartDirector in the XYChart.
 *  \param C Coordinates To Be Translated/Translated (Input & Output)
 *  \return Result of the Action (Success/Failure)
 *  \author Chibiao Chen
 */
bool TranslateCoordinates(DBL_VCTR_2D& C)
{
	int N = int(C.size());
	if (N == 0 || C[0].size() != 2)
	{
		return false;
	}
	double xMin = +1.0E+300;
	double xMax = -1.0E+300;
	double yMin = +1.0E+300;
	double yMax = -1.0E+300;
	int i;
	for (i = 0; i < N; i++)
	{
		if (C[i][0] < xMin)
		{
			xMin = C[i][0];
		}
		if (C[i][0] > xMax)
		{
			xMax = C[i][0];
		}
		if (C[i][1] < yMin)
		{
			yMin = C[i][1];
		}
		if (C[i][1] > yMax)
		{
			yMax = C[i][1];
		}
	}
	double xOffset = -xMin + (xMax - xMin) * 2.0;
	double yOffset = -yMin + (yMax - yMin) * 2.0;
	for (i = 0; i < N; i++)
	{
		C[i][0] += xOffset;
		C[i][1] += yOffset;
	}

	return true;
}

//! The TESS MCMC Algorithms
int main(int argc, char* argv[])
{
	gsl_set_error_handler_off();
	
	// Required Parameters
	string F;			// Name of Data File
	int N;				// Number of Samples/Individuals
	int A;				// Haploid, Diploid, ...
	int L;				// Number of Loci
	INT_VCTR_1D J;		// Number of Distinct Alleles Observed at Every Locus
	int K;				// Number of Cluster
	double psi;			// Interaction Parameter of HMRF
	double lambda;		// Parameter of Dirichlet Allele Frequency Model
	int TotalSweep;		// Total Number of Sweeps
	int BurnIn;			// Burn In Number of Sweeps

	// Optional Parameters
	int ER = 0;			// Number of Extra Rows
	int EC = 0;			// Number of Extra Columns
	int nd = 0;			// Number of Dummy Individuals
	string FoI;			// Folder of Input Data File
	string FoO;			// Folder of Output Result Files
	string CF;			// Name of CF File
	string Voronoi;			// Name of Voronoi File
	bool nj = false;		// Init with neighbor joining tree ?
	bool specialFormat = false;     // Special format
	string geoFile;			// Name of geographic distances file
	string scOpt = "m";		// How to initialize the scale parameter theta. Default: mean distance
	bool scSpecified = false;	// Has user specified a value for the scale parameter. Default: no, use the mean distance
	double theta = 1.0;		// Geographic scale parameter
	
	double minDIC = 1.7e+308;	// min DIC of all runs
	double MLpsi;			// Spatial interaction parameter of best run
	
	bool bF = false;
	bool bN = false;
	bool bA = false;
	bool bL = false;
	bool bK = false;
	bool bP = false;
	bool bD = false;
	bool bS = false;
	bool bB = false;

	if (argc < 10)
	{
		ShowUsage();
		exit(EXIT_FAILURE);
	}
	else
	{
		// Parse the command line parameters.
		for(int ac = 1; ac < argc; ac++)
		{
			switch (argv[ac][1])
			{
				case 'F':
					F = (const char*) &argv[ac][2];
					bF = true;
					break;

				case 'N':
					N = atoi((const char*) &argv[ac][2]);
					bN = true;
					break;

				case 'A':
					A = atoi((const char*) &argv[ac][2]);
					bA = true;
					break;

				case 'L':
					L = atoi((const char*) &argv[ac][2]);
					bL = true;
					break;

				case 'K':
					K = atoi((const char*) &argv[ac][2]);
					bK = true;
					break;

				case 'P':
					psi = atof((const char*) &argv[ac][2]);
					bP = true;
					break;

				case 'D':
					lambda = atof((const char*) &argv[ac][2]);
					bD = true;
					break;

				case 'S':
					TotalSweep = atoi((const char*) &argv[ac][2]);
					bS = true;
					break;

				case 'B':
					BurnIn = atoi((const char*) &argv[ac][2]);
					bB = true;
					break;

				case 'r':
					ER = atoi((const char*) &argv[ac][2]);
					break;

				case 'c':
					EC = atoi((const char*) &argv[ac][2]);
					break;

				case 'd':
					nd = atoi((const char*) &argv[ac][2]);
					break;

				case 'i':
					FoI = (const char*) &argv[ac][2];
					break;
					
				case 'j':
					if (argv[ac][2] != 'y' && argv[ac][2] != 'n')
					{
						cerr << "Unknown Option - " << argv[ac] << endl;
						ShowUsage();
						exit(EXIT_FAILURE);
					}
					if (argv[ac][2] == 'y')
					{
						nj = true;
					}
					if (argv[ac][2] == 'n')
					{
						nj = false;
					}
					break;
					
				case 'g':
					geoFile = (const char*) &argv[ac][2];
					break;

				case 'o':
					FoO = (const char*) &argv[ac][2];
					break;

				case 'p':
					CF = (const char*) &argv[ac][2];
					break;
					
				case 's':
					if (argv[ac][2] == 'c')
					{
						if (argv[ac][3] == 'm' || argv[ac][3] == 'd' || argv[ac][3] == 'x')
						{
							scOpt = argv[ac][3];
						}
						else 
						{
							theta= atof((const char*) &argv[ac][3]);	
							scSpecified = true;
						}
						break;
					}
					
					if (argv[ac][2] == 'p')
					{
						//SPecial data format
						if (argv[ac][3] != 'y' && argv[ac][3] != 'n')
						{
							cerr << "Unknown Option - " << argv[ac] << endl;
							ShowUsage();
							exit(EXIT_FAILURE);
						}
						if (argv[ac][3] == 'y')
						{
							specialFormat = true;
						}
						if (argv[ac][3] == 'n')
						{
							specialFormat = false;
						}
						break;
					}
					else
					{
						cerr << "Unknown Option - " << argv[ac] << endl;
						ShowUsage();
						exit(EXIT_FAILURE);
						break;
					}

				case 'v':
					Voronoi = (const char*) &argv[ac][2];
					break;

				default:  // Unknown parameters.
					cerr << "Unknown Option - " << argv[ac] << endl;
					ShowUsage();
					exit(EXIT_FAILURE);
			}
		}
	}

	if (!bF || !bN || !bA || !bL || !bK || !bP || !bD || !bS || !bB)
	{
		ShowUsage();
		exit(EXIT_FAILURE);
	}

	string FIFN = !FoI.empty() ? FoI + "/" + F : F;
	string FOFN = !FoO.empty() ? FoO + "/" + F : F;

	int NumOfDigits = (int) log10((double) TotalSweep) + 1;

	// Loop Index Variables
	int i, a, l, j, k;

	DBL_VCTR_2D C;
	InitVctr2D(C, N + nd, 2, 0.0);

	INT_VCTR_3D X;
	InitVctr3D(X, N, A, L, 0);  // L Loci
	


	LoadData(C, X, ER, EC, FIFN, specialFormat);
	TranslateCoordinates(C);
	
	//geographic file available ?
	bool hasGeo = !geoFile.empty();
	
	DBL_VCTR_2D geo; //the geographic distances matrix
	if (hasGeo)
	{
		InitVctr2D(geo,N,N,0.0);
		LoadVector2D(geo,geoFile);
		if (!scSpecified)
		{
			computeSymmMatrixStat(theta,geo,scOpt);
		}
// 		cout << "scOpt = " << scOpt << " theta = " << theta << endl;
	}

	vector< SPoint > vptC;
	vector< bool > vbKeep;
	vector< double > vdDataMinMax;
	vector< double > vdDataScale;
	vector< CEdge > veVoronoi;
	INT_VCTR_2D vvnNS;

	
	//Allocation of seed
	const gsl_rng_type *T;
	gsl_rng_env_setup();
	T = gsl_rng_default;
	gsl_rng* rng = NULL;
	AllocRNG(rng, T);
	
	
	if (Voronoi == "" ||
		!CVoronoi::LoadVoronoi(vptC, vbKeep, vdDataMinMax, vdDataScale,
			veVoronoi, vvnNS, Voronoi))
	{
		
		int nbDigitMax = 0;
		int nbDigit;
		string xCoord = "";
		string yCoord = "";
		
		for (i = 0; i < N; i ++)
		{
			stringstream tmp1,tmp2;
			tmp1 << C[i][0];
			xCoord = tmp1.str();
			tmp2 << C[i][1];
			yCoord = tmp2.str();
			nbDigit = max((xCoord.size()-xCoord.find('.')),(yCoord.size()-yCoord.find('.'))) - 1;

			if (nbDigit > nbDigitMax)
			{
				nbDigitMax = nbDigit;
			}
			tmp1.clear();
			tmp2.clear();
		}
	

	
		double perturb = gsl_pow_int(10,nbDigitMax);
		perturb = 1/(10.0*perturb);
		
				
		for (i = 0; i < N; i++)
		{
			C[i][0] += gsl_ran_gaussian(rng,perturb);
			C[i][1] += gsl_ran_gaussian(rng,perturb);
		}
		
		cout << "Generating Voronoi Tessellation & Neighborhood System... ";
		vptC.clear();
		SPoint ptThis;
		for (i = 0; i < N + nd; i++)
		{
			vptC.push_back(ptThis);
			vptC[i].m_nIndex = i;
			vptC[i].m_psXY.m_dX = C[i][0];
			vptC[i].m_psXY.m_dY = C[i][1];
		}
		CVoronoi::GetVoronoi(vptC, vbKeep, vdDataMinMax, vdDataScale, veVoronoi);
		string strVD = FIFN + "VD.png";
		CVoronoi::DrawVoronoi(vptC, N, vbKeep, vdDataMinMax, veVoronoi, false, strVD);
		CVoronoi::GetNeighborhoodSystem(N, veVoronoi, vvnNS);
		string strND = FIFN + "ND.png";
		CVoronoi::DrawNeighborhoodSystem(vptC, vdDataScale, vvnNS, true, strND);
		string strVC = FIFN + "VC";
		CVoronoi::SaveVoronoi(vptC, vbKeep,
		vdDataMinMax, vdDataScale, veVoronoi, vvnNS, strVC);
		cout << "Done" << endl;
	}
	/*
	string NSstr = "vvnNS.txt";
	CVoronoi::WriteNStoMyFile(vvnNS,NSstr);*/

	// Determine the actual J from the data X.
	int LN = L;
	

	vector< set< int > > alleleS;
	set< int > emptyS;
	for (l = 0; l < LN; l++)
	{
		alleleS.push_back(emptyS);
	}
	for (l = 0; l < LN; l++)
	{
		for (i = 0; i < N; i++)
		{
			for (a = 0; a < A; a++)
			{
				alleleS[l].insert(X[i][a][l]);
			}
		}
	}

	INT_VCTR_2D alleleV;
	INT_VCTR_1D emptyV;
	for (l = 0; l < LN; l++)
	{
		alleleV.push_back(emptyV);
	}
	
	for (l = 0; l < LN; l++)
	{
		set< int >::const_iterator csiB = alleleS[l].begin();
		set< int >::const_iterator csiE = alleleS[l].end();
		// Discard the missing allele.
		// Missing allele is represented by a negative integer.
		if (*csiB < 0)
		{
			++csiB;
		}
		set< int >::const_iterator csi;
		for (csi = csiB; csi != csiE; ++csi)
		{
			alleleV[l].push_back(*csi);
		}
	}

	J.clear();
	for (l = 0; l < LN; l++)
	{
		J.push_back(static_cast<int>(alleleV[l].size()));
	}
	int JMax = *max_element(J.begin(), J.end());


	clock_t s_clock = clock();

	

	DBL_VCTR_2D DirParams;
	InitVctr2D(DirParams, L, J, lambda);

	CSweepParams spSweep;
	InitVctr1D(spSweep.z, N, 0);
	InitVctr3D(spSweep.p, K, L, J, 0.0);

	
	//Init Z, P with NJ tree ? (individuals level)
	if (nj)
	{
		INT_VCTR_1D InitLabels = HClustering(X, WAR, K, true);
		for (i = 0; i < N; i++)
		{
			spSweep.z[i]=InitLabels[i];
		}
	

		INT_VCTR_1D OI(N, 0);  
		for (i = 0; i < N; i++)
		{
			OI[i] = i;
		}
		int SN;
		if (L < 120)
		{
			SN = static_cast<int>(N * 0.25);
		}
		else
		{
			SN = static_cast<int>(N * 0.05);
		}
		if (SN == 0)
		{
			SN = 1;
		}
		INT_VCTR_1D SI(SN, 0);
		gsl_ran_choose(rng, &SI[0], SN, &OI[0], N, sizeof(int));
		INT_VCTR_1D KI(K, 0);
		for (k = 0; k < K; k++)
		{
			KI[k] = k;
		}
		for (i = 0; i < SN; i++)
		{
			int knew;
			do 
			{
				gsl_ran_choose(rng, &knew, 1, &KI[0], K, sizeof(int));
			} 
			while (knew == InitLabels[SI[i]]);
			InitLabels[SI[i]] = knew;
		}
		for (i = 0; i < N; i++)
		{

			spSweep.z[i]=InitLabels[i];
		}
			
		//now let's initialize p
		for (k = 0; k < K; k++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					int nSum = 0;
					int nInd = 0;
					for (i = 0; i < N; i++)
					{
						for (a = 0; a < A; a++)
						{
							if (spSweep.z[i] == k)
							{
								nInd++;
								if (X[i][a][l] == alleleV[l][j])
								{
									nSum++;
								}
							}
						}
					}
					if (nInd == 0) 
					{
						spSweep.p[k][l][j] = 0.0;
					}
					else
					{
						spSweep.p[k][l][j] = (double)(nSum/(double)(nInd));
					}
				}
			}
		}
		string strIM = FOFN + "IM.png";
		PlotHC(strIM.c_str(), vptC, vbKeep, vdDataMinMax, veVoronoi, spSweep.z);
	}
	
	
	// Initialize z and p using configuration file or priors.
	if (CF == "" || !LoadCF(spSweep.z, spSweep.p, CF))
	{
		if (!nj)
		{
			for (i = 0; i < N; i++)
			{
				spSweep.z[i] = gsl_rng_uniform_int(rng, K);
			}
			for (k = 0; k < K; k++)
			{
				for (l = 0; l < L; l++)
				{
					gsl_ran_dirichlet(rng, J[l], &DirParams[l][0], &spSweep.p[k][l][0]);
				}
			}
		}
	}

	if (!nj)
	{
		string strIM = FOFN + "IM.png";
		PlotHC(strIM.c_str(), vptC, vbKeep, vdDataMinMax, veVoronoi, spSweep.z);
	}

	// Initialize auxiliary variable n.
	INT_VCTR_3D n;
	InitVctr3D(n, K, L, J, 0);

	// Initialize auxiliary variable logp.
	DBL_VCTR_3D logp;
	InitVctr3D(logp, K, L, J, 0.0);

	// Initialize clustering probabilities cpSweep and cpAverage.
	DBL_VCTR_2D cpAverage;
	InitVctr2D(cpAverage, N, K, 0.0);

	CSweepParams spAverage;
	InitVctr1D(spAverage.z, N, 0);
	InitVctr3D(spAverage.p, K, L, J, 0.0);

	DBL_VCTR_1D LH(TotalSweep, 0.0);  // Log-Likelihood History

	

	// Now, let's start our Markov chain Monte Carlo.
	int KeptSweep = TotalSweep - BurnIn;
	for (int s = 0; s < TotalSweep; s++)
	{
		// Update p.
		for (k = 0; k < K; k++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					int nSum = 0;
					for (i = 0; i < N; i++)
					{
						for (a = 0; a < A; a++)
						{
							if (X[i][a][l] == alleleV[l][j] && spSweep.z[i] == k)
							{
								nSum++;
							}
						}
					}
					n[k][l][j] = nSum;
				}
			}
		}
		for (k = 0; k < K; k++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					DirParams[l][j] = n[k][l][j] + lambda;
				}
				gsl_ran_dirichlet(rng, J[l], &DirParams[l][0], &spSweep.p[k][l][0]);
			}
		}

		// Update logp.
		for (k = 0; k < K; k++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					logp[k][l][j] = log(spSweep.p[k][l][j]);
				}
			}
		}

		// Update z.
		for (i = 0; i < N; i++)
		{
			int kOld = spSweep.z[i];
			int kNew = gsl_rng_uniform_int(rng, K);
			if (kNew != kOld)
			{
				double dLogRatio = 0.0;
				for (a = 0; a < A; a++)
				{
					for (l = 0; l < L; l++)
					{
						for (j = 0; j < J[l]; j++)
						{
							if (X[i][a][l] == alleleV[l][j])
							{
								dLogRatio += logp[kNew][l][j] - logp[kOld][l][j];
								break;
							}
						}
					}
				}
				
				double nDelta = 0.0;
				for (j = 0; j < static_cast<int>(vvnNS[i].size()); j++)
				{
					if (spSweep.z[vvnNS[i][j]] == kNew)
					{
						if (!hasGeo)
						{
							nDelta+=1.0;
						}
						else
						{
							nDelta+=exp(-geo[i][vvnNS[i][j]]/theta);
						}
					}
					if (spSweep.z[vvnNS[i][j]] == kOld)
					{
						if (!hasGeo)
						{
							nDelta-=1.0;
						}
						else
						{
							nDelta-=exp(-geo[i][vvnNS[i][j]]/theta);
						}
					}
				}

				double u = gsl_rng_uniform(rng);
				if (u < exp(dLogRatio + psi * nDelta))
				{
					spSweep.z[i] = kNew;
				}
			}
		}

		// Update Log-Likelihood.
		double LCurr = 0.0;
		for (i = 0; i < N; i++)
		{
			for (a = 0; a < A; a++)
			{
				for (l = 0; l < L; l++)
				{
					for (j = 0; j < J[l]; j++)
					{
						if (X[i][a][l] == alleleV[l][j])
						{
							LCurr += logp[spSweep.z[i]][l][j];
							break;
						}
					}
				}
			}
		}
		LH[s] = LCurr;

		// Accumulate results after burn-in.
		if (s > BurnIn - 1)
		{
			// p
			for (k = 0; k < K; k++)
			{
				for (l = 0; l < L; l++)
				{
					for (j = 0; j < J[l]; j++)
					{
						spAverage.p[k][l][j] += spSweep.p[k][l][j];
					}
				}
			}

			// cp
			for (i = 0; i < N; i++)
			{
				for (k = 0; k < K; k++)
				{
					if (spSweep.z[i] == k)
					{
						cpAverage[i][k] += 1.0;
						break;
					}
				}
			}
		}

		int spr = s + 1;
		if (spr % 100 == 0)
		{
			cout.width(NumOfDigits);
			cout.fill('0');
			cout << spr << " sweeps completed." << endl;
		}
	}

	string strLH = FOFN + "LH.png";
	PlotLH(strLH.c_str(), LH);

	// Calculate the Monte Carlo averages after MCMC.
	// p
	for (k = 0; k < K; k++)
	{
		for (l = 0; l < L; l++)
		{
			for (j = 0; j < J[l]; j++)
			{
				spAverage.p[k][l][j] /= KeptSweep;
			}
		}
	}

	// cp
	for (i = 0; i < N; i++)
	{
		for (k = 0; k < K; k++)
		{
			cpAverage[i][k] /= KeptSweep;
		}
	}

	// Hard Clustering
	INT_VCTR_1D hc(N, 0);
	for (i = 0; i < N; i++)
	{
		double dMax = -1.0;
		for (k = 0; k < K; k++)
		{
			if (cpAverage[i][k] > dMax)
			{
				dMax = cpAverage[i][k];
				hc[i] = k;
			}
		}
	}
	string strHC = FOFN + "HC.png";
	PlotHC(strHC.c_str(), vptC, vbKeep, vdDataMinMax, veVoronoi, hc);

// 	string strCF = FIFN + "CF";
// 	SaveCF(hc, spAverage.p, strCF);

	// pi
	DBL_VCTR_1D pi(K, 0.0);
	for (k = 0; k < K; k++)
	{
		for (i = 0; i < N; i++)
		{
			pi[k] += cpAverage[i][k];
		}
	}
	for (k = 0; k < K; k++)
	{
		pi[k] /= N;
	}

	// Log the results to a text file.
	string strTR = FOFN + "TR.txt";
	ofstream ofsTR(strTR.c_str());

	// Estimated Clustering Probabilities for Samples
	ofsTR << "Estimated Clustering Probabilities for Samples" << endl;
	ofsTR << "Sample\t";
	for (k = 0; k < K; k++)
	{
		ofsTR << "***Cluster  ";
		ofsTR.width(2);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << k + 1 << "***\t";
	}
	ofsTR << endl;
	for (i = 0; i < N; i++)
	{
		ofsTR.width(6);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << i + 1<<":"<< "\t";
		for (k = 0; k < K; k++)
		{
			ofsTR.width(17);
			ofsTR.precision(15);
			ofsTR << fixed;
			ofsTR << cpAverage[i][k] << "\t";
		}
		ofsTR << "\t" << "1";
		ofsTR << endl;
	}
	ofsTR << endl << endl << endl;

	// Estimated Allele Frequencies
	ofsTR << "Estimated Allele Frequencies" << endl;
	for (k = 0; k < K; k++)
	{
		ofsTR << "Cluster ";
		ofsTR.width(2);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << k + 1 << endl;
		for (l = 0; l < L; l++)
		{
			ofsTR << "Locus ";
			ofsTR.width(3);
			ofsTR.fill('0');
			// Make sure the first index is 1 not 0.
			ofsTR << l + 1 << endl;
			for (j = 0; j < J[l]; j++)
			{
				ofsTR << "Allele ";
				ofsTR.width(3);
				ofsTR.fill('0');
				ofsTR << alleleV[l][j] << "\t";
				ofsTR.width(17);
				ofsTR.precision(15);
				ofsTR << fixed;
				ofsTR << spAverage.p[k][l][j] << endl;
			}
			ofsTR << endl;
		}
		ofsTR << endl;
	}
	ofsTR << endl;

	// Estimated Mixing Proportions
	ofsTR << "Estimated Mixing Proportions" << endl;
	for (k = 0; k < K; k++)
	{
		ofsTR.width(17);
		ofsTR.precision(15);
		ofsTR << fixed;
		ofsTR << pi[k] << endl;
	}
	ofsTR << endl << endl << endl;

	
	ofsTR.close();

	string strAP = FOFN + "AP.png";
	PlotAP(strAP.c_str(), cpAverage);

	// Calculate Average Log-Likelihood
	int NoS = int(0.1 * KeptSweep);  // Number of Sweeps Used in Calculation
	if (NoS == 0)
	{
		NoS = 1;
	}
	double dAL = 0.0;
	for (i = TotalSweep - NoS; i < TotalSweep; i++)
	{
		dAL += LH[i];
	}
	dAL /= NoS;
	string strAL = FOFN + "AL.txt";
	ofstream ofsAL(strAL.c_str());
	ofsAL << dAL << endl;
	ofsAL.close();
		
	
	// Calculate DIC
	double dALdic = 0.0;
	for (i = BurnIn; i < TotalSweep; i ++)
	{
		dALdic += LH[i];
	}
	dALdic /= KeptSweep;

	
	double tmpDIC = 0.0;
	double tmpDIC2 = 0.0;
	double tmp = 0.0;
	for (i = 0; i < N; i++)
	{
		for (a = 0; a < A; a++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					if (X[i][a][l] == alleleV[l][j])
					{
						tmp = 0.0;
						for (k = 0; k < K; k ++)
						{
							tmp+=spAverage.p[k][l][j]*cpAverage[i][k];
						}
						tmpDIC += log(tmp);
						tmpDIC2 += log(spAverage.p[hc[i]][l][j]);
						break;
					}
				}
			}
		}
	}
	
	double pd = -2*dALdic + 2*tmpDIC;
	double pd2 = -2*dALdic + 2*tmpDIC2;
	double DIC = pd - 2*dALdic;
	double DIC2 = pd2 - 2*dALdic;
	
	string strPD = FOFN + "PD.txt";
	ofstream ofsPD(strPD.c_str());
	ofsPD << pd << endl;
	ofsPD.close();
	
	string strPD2 = FOFN + "PD2.txt";
	ofstream ofsPD2(strPD2.c_str());
	ofsPD2 << pd2 << endl;
	ofsPD2.close();
	
	string strDIC = FOFN + "DIC.txt";
	ofstream ofsDIC(strDIC.c_str());
	ofsDIC << DIC << endl;
	ofsDIC.close();
	
	string strDIC2 = FOFN + "DIC2.txt";
	ofstream ofsDIC2(strDIC2.c_str());
	ofsDIC2 << DIC2 << endl;
	ofsDIC2.close();

	//Store the min DIC on all runs
        //Restore the previous max likelihood
	
	string strML = FIFN + "MDIC_noAdm";
	if (LoadML(MLpsi,minDIC,strML,true))
	{
		//A previous run has been saved. Have we done better ?
		if (DIC < minDIC)
		{
			//Yes, we did better. Saving the new DIC and the state reached by the markov chain.
			SaveML(psi,DIC,strML);
			string strCF = FIFN + "CFnoAdm";
			SaveCF(hc, spAverage.p,strCF);
		}
	} 
	else
	{
		//No maxlikelihood has been stored until now. We save the DIC and the state reached by the markov chain.
		SaveML(psi,DIC,strML);
		string strCF = FIFN + "CFnoAdm";
		SaveCF(hc,spAverage.p,strCF);
	}
	
	
	FreeRNG(rng);

	clock_t e_clock = clock();

	cout << fixed;
	cout.precision(3);
	cout << "This run takes " << static_cast<double>(e_clock - s_clock) /
		CLOCKS_PER_SEC << " seconds to complete on this computer." << endl;
	return 0;
}
